#!/usr/bin/env python3
"""Gets packed pixels from a gimp image.

This represents a single level in an imageHierarchy
"""
from __future__ import annotations

import math
import zlib

import PIL.Image
from PIL.Image import Image

from .GimpIOBase import IO, GimpIOBase


class GimpImageLevel(GimpIOBase):
	"""Gets packed pixels from a gimp image.

	This represents a single level in an imageHierarchy
	"""

	def __init__(self, parent):
		GimpIOBase.__init__(self, parent)
		self.width = 0
		self.height = 0
		self._tiles = None  # tile PIL images
		self._image = None

	def decode(self, data: bytes, index: int = 0):
		"""Decode a byte buffer.

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		ioBuf = IO(data, index)
		# print 'Decoding image level at',ioBuf.index
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		if self.width != self.parent.width or self.height != self.parent.height:
			currentSize = "(" + str(self.width) + "," + str(self.height) + ")"
			expectedSize = "(" + str(self.parent.width) + "," + str(self.parent.height) + ")"
			msg = " Usually this implies file corruption."
			raise Exception("Image data size mismatch. " + currentSize + "!=" + expectedSize + msg)
		self._tiles = []
		self._image = None
		for y in range(0, self.height, 64):
			for x in range(0, self.width, 64):
				ptr = self._pointerDecode(ioBuf)
				size = (min(self.width - x, 64), min(self.height - y, 64))
				totalBytes = size[0] * size[1] * self.bpp
				if self.doc.compression == 0:  # none
					data = ioBuf.data[ptr : ptr + totalBytes]
				elif self.doc.compression == 1:  # RLE
					data = self._decodeRLE(ioBuf.data, size[0] * size[1], self.bpp, ptr)
				elif self.doc.compression == 2:  # zip
					data = zlib.decompress(
						ioBuf.data[ptr : ptr + totalBytes + 24]
					)  # guess how many bytes are needed
				else:
					raise Exception(
						"ERR: unsupported compression mode " + str(self.doc.compression)
					)
				subImage = PIL.Image.frombytes(self.mode, size, bytes(data), decoder_name="raw")
				self._tiles.append(subImage)
		_ = self._pointerDecode(ioBuf)  # list ends with nul character
		return ioBuf.index

	def encode(self):
		"""Encode this object to a byte buffer."""
		dataioBuf = IO()
		ioBuf = IO()
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		dataIndex = ioBuf.index + self.POINTER_SIZE * (len(self.tiles) + 1)
		for tile in self.tiles:
			ioBuf.addBytes(self._pointerEncode(dataIndex + dataioBuf.index))
			data = tile.tobytes()
			if self.doc.compression == 0:  # none
				pass
			elif self.doc.compression == 1:  # RLE
				data = self._encodeRLE(data, self.bpp)
				# raise Exception('RLE Compression is a work in progress!')
			elif self.doc.compression == 2:  # zip
				data = zlib.compress(data)
			else:
				raise Exception("ERR: unsupported compression mode " + str(self.doc.compression))
			dataioBuf.addBytes(data)
		ioBuf.addBytes(self._pointerEncode(0))
		ioBuf.addBytes(dataioBuf.data)
		return ioBuf.data

	def _decodeRLE(self, data, pixels, bpp, index=0):
		"""Decode RLE encoded image data."""
		ret = [[] for chan in range(bpp)]
		for chan in range(bpp):
			n = 0
			while n < pixels:
				opcode = data[index]
				index += 1
				if 0 <= opcode <= 126:  # a short run of identical bytes
					val = data[index]
					index += 1
					for _ in range(opcode + 1):
						ret[chan].append(val)
						n += 1
				elif opcode == 127:  # A long run of identical bytes
					m = data[index]
					index += 1
					b = data[index]
					index += 1
					val = data[index]
					index += 1
					amt = m * 256 + b
					for _ in range(amt):
						ret[chan].append(val)
						n += 1
				elif opcode == 128:  # A long run of different bytes
					m = data[index]
					index += 1
					b = data[index]
					index += 1
					amt = m * 256 + b
					for _ in range(amt):
						val = data[index]
						index += 1
						ret[chan].append(val)
						n += 1
				elif 129 <= opcode <= 255:  # a short run of different bytes
					amt = 256 - opcode
					for _ in range(amt):
						val = data[index]
						index += 1
						ret[chan].append(val)
						n += 1
				else:
					print("Unreachable branch", opcode)
					raise Exception()
		# flatten/weave the individual channels into one strream
		flat = bytearray()
		for i in range(pixels):
			for chan in range(bpp):
				flat.append(ret[chan][i])
		return flat

	def _encodeRLE(self, data, bpp):
		"""Encode image to RLE image data."""

		def countSame(data, startIdx):
			"""Count how many times bytes are identical."""
			idx = startIdx
			l = len(data)
			if idx >= l:
				return 0
			c = data[idx]
			idx = startIdx + 1
			while idx < l and data[idx] == c:
				idx += 1
			return idx - startIdx

		def countDifferent(data, startIdx):
			"""Count how many times bytes are different."""
			idx = startIdx
			l = len(data)
			if idx >= l:
				return 1
			c = data[idx]
			idx = startIdx + 1
			while idx < (l - 1) and data[idx] != c:
				idx += 1
				c = data[idx]
			return idx - startIdx

		def rleEncodeChan(data):
			"""Rle encode a single channel of data."""
			ret = []
			idx = 0
			while idx < len(data):
				nRepeats = countSame(data, 0)
				if nRepeats == 1:  # different bytes
					nDifferences = countDifferent(data, 1)
					if nDifferences <= 127:  # short run of different bytes
						ret.append(129 + nRepeats - 1)
						ret.append(data[idx])
						idx += nDifferences
					else:  # long run of different bytes
						ret.append(128)
						ret.append(math.floor(nDifferences / 256.0))
						ret.append(nDifferences % 256)
						ret.append(data[idx])
						idx += nDifferences
				elif nRepeats <= 127:  # short run of same bytes
					ret.append(nRepeats - 1)
					ret.append(data[idx])
					idx += nRepeats
				else:  # long run of same bytes
					ret.append(127)
					ret.append(math.floor(nRepeats / 256.0))
					ret.append(nRepeats % 256)
					ret.append(data[idx])
					idx += nRepeats
			return ret

		# if there is only one channel, encode and return it directly
		if bpp == 1:
			return rleEncodeChan(data)
		# split into channels
		dataByChannel = []
		for chan in range(bpp):
			chanData = []
			for index in range(chan, bpp, len(data)):
				chanData.append(data[index])
			dataByChannel.append(chanData)
		# encode each channel
		for index in range(len(dataByChannel)):  # iterate through 2d array
			dataByChannel[index] = rleEncodeChan(dataByChannel[index])
		# join and return
		return "".join("".join(str(x)) for x in dataByChannel)

	@property
	def bpp(self):
		"""Get bpp."""
		return self.parent.bpp

	@property
	def mode(self):
		"""Get mode."""
		MODES = [None, "L", "LA", "RGB", "RGBA"]
		return MODES[self.bpp]

	@property
	def tiles(self):
		"""Get tiles."""
		if self._tiles is not None:
			return self._tiles
		if self.image is not None:
			return self._imgToTiles(self.image)
		return None

	def _imgToTiles(self, image):
		"""
		break an image into a series of tiles, each<=64x64
		"""
		ret = []
		for y in range(0, self.height, 64):
			for x in range(0, self.width, 64):
				bounds = (x, y, min(self.width - x, 64), min(self.height - y, 64))
				ret.append(image.crop(bounds))
		return ret

	@property
	def image(self) -> Image:
		"""
		Get a final, compiled image
		"""
		if self._image is None:
			self._image = PIL.Image.new(self.mode, (self.width, self.height), color=None)
			tileNum = 0
			for y in range(0, self.height, 64):
				for x in range(0, self.width, 64):
					subImage = self._tiles[tileNum]
					tileNum += 1
					self._image.paste(subImage, (x, y))
			# self._tiles = None
		return self._image

	@image.setter
	def image(self, image: Image):
		self._image = image
		self._tiles = None
		self.width = image.width
		self.height = image.height
		self.tiles = None

	def __repr__(self, indent: str = ""):
		"""Get a textual representation of this object."""
		ret = []
		ret.append("Size: " + str(self.width) + " x " + str(self.height))
		return indent + (("\n" + indent).join(ret))



#!/usr/bin/env python3
"""Pure python implementation of the gimp gtp tool preset format.
"""
from __future__ import annotations

from io import BytesIO

import brackettree


class ParenFileValue:
	"""A parentheses-based file format.

	(possibly "scheme" language?)
	"""

	def __init__(self, name: str = None, value: str = "", children=None):
		"""parenthesis based file

		Args:
			name (str, optional): name of the file. Defaults to None.
			value (str, optional): some value, str(float) or str. Defaults to "".
			children ([type], optional): children. Defaults to None.
		"""

		self.name = name
		try:
			float(value)
			self.value = value
		except ValueError:
			self.value = '"' + value + '"'
		if value in ("yes", "no"):
			self.value = value
		self.children = children

	def _addValue(self, bufArray):
		if self.name is None:  # first value is the name
			self.name = bufArray
		if bufArray:
			bufArray = "".join(bufArray)
			if bufArray in ("yes", "no"):  # boolean
				self.value.append(bufArray == "yes")
			elif bufArray[0].isdigit():
				self.value.append(float(bufArray))
			elif bufArray[0] == '"':
				self.value.append(bufArray[1:-1])
			else:
				raise Exception('What kind of value is "' + bufArray + '"?')

	def __repr__(self):
		"""Get a textual representation of this object."""
		ret = []
		ret.append("(" + str(self.name))
		ret.append(" " + self.value)
		if self.children is None:
			ret.append(")")
		ret.append("\n")
		if self.children is not None:
			for index, child in enumerate(self.children):
				ret.append("    (" + str(child.name))
				ret.append(" " + child.value + ")")
				if index == len(self.children) - 1:
					ret.append(")")
				ret.append("\n")
		return "".join(ret)


def parenFileDecode(data):
	"""Decode a parentheses-based file format.

	(possibly "scheme" language?)
	"""
	nodes = brackettree.Node(data)
	values = walkTree(nodes.items)
	return values


def walkTree(items):
	"""Walk the tree."""
	values = []
	for item in items:
		if isinstance(item, brackettree.RoundNode):
			if len(item.items) == 1:
				values.append(
					ParenFileValue(
						item.items[0].split(" ")[0].strip(), item.items[0].split(" ")[1].strip()
					)
				)
			if len(item.items) == 2:
				values.append(ParenFileValue(item.items[0].strip(), item.items[1].items[0].strip()))
			elif len(item.items) > 2:
				values.append(
					ParenFileValue(
						item.items[0].strip(),
						item.items[1].items[0].strip(),
						walkTree(item.items[2:]),
					)
				)
	return values


def parenFileEncode(values):
	"""Encode a values tree to a buffer."""
	ret = []
	ret.append("# GIMP tool preset file\n\n")
	ret.append("")
	for val in values:
		if val.name is not None:
			ret.append(str(val))
	ret.append("")
	ret.append("\n# end of GIMP tool preset file\n")
	ret.append("")
	return "".join(ret)


class GimpGtpToolPreset:
	"""Pure python implementation of the gimp gtp tool preset format."""

	def __init__(self, fileName=None):
		self.fileName = None
		self.values = []
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		if hasattr(fileName, "read"):
			self.fileName = fileName.name
			f = fileName
		else:
			self.fileName = fileName
			f = open(fileName)
		data = f.read()
		self.decode(data)
		f.close()

	def decode(self, data, index=0):
		"""Decode a byte buffer.

		:param data: data buffer to decode
		"""
		self.values = parenFileDecode(data)
		return index

	def encode(self):
		"""Encode to a byte array."""
		return parenFileEncode(self.values).encode("utf-8")

	def save(self, tofileName=None, toExtension=None):
		"""Save this gimp tool preset to a file."""
		if toExtension is None:
			if tofileName is not None:
				toExtension = tofileName.rsplit(".", 1)
				if len(toExtension) > 1:
					toExtension = toExtension[-1]
				else:
					toExtension = None
		if hasattr(tofileName, "write"):
			file = tofileName
		else:
			file = open(tofileName, "wb")
		file.write(self.encode())
		file.close()

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		for value in self.values:
			ret.append(value.__repr__(indent + "\t"))
		return "\n".join(ret)



#!/usr/bin/env python3
"""Gimp color gradient.
"""
from __future__ import annotations

from io import BytesIO


class GradientSegment:
	"""Single segment within a gradient."""

	BLEND_FUNCTIONS = [
		"linear",
		"curved",
		"sinusoidal",
		"spherical (increasing)",
		"spherical (decreasing)",
		"step",
	]
	COLOR_TYPES = ["RGB", "HSV CCW", "HSV CW"]
	ENDPOINT_COLOR_TYPES = [
		"fixed",
		"foreground",
		"foreground transparent",
		"background",
		"background transparent",
	]

	def __init__(self):
		"""Single segment within a gradient."""
		self.leftPosition = 0
		self.middlePosition = 0.5
		self.rightPosition = 1.0
		self.leftColor = (0, 0, 0, 0)
		self.rightColor = (255, 255, 255, 0)
		self.blendFunc = None  # one of self.BLEND_FUNCTIONS
		self.colorType = None  # one of self.COLOR_TYPES
		self.leftColorType = None  # one of self.ENDPOINT_COLOR_TYPES
		self.rightColorType = None  # one of self.ENDPOINT_COLOR_TYPES

	def getColor(self, percent):
		"""Given a decimal percent (1.0 = 100%) retrieve the appropriate...

		color for this point in the gradient.
		"""
		raise NotImplementedError()

	def decode(self, dataIn: str):
		"""Decode a byte buffer.

		Args:
			dataIn (str): data buffer to decode

		Raises:
			Exception: [description]
		"""
		data = dataIn.split(" ")
		if len(data) < 11 or len(data) > 15:
			raise Exception("Data table is unexpected size. " + str(len(data)))
		self.leftPosition = float(data[0])
		self.middlePosition = float(data[1])
		self.rightPosition = float(data[2])
		self.leftColor = (float(data[3]), float(data[4]), float(data[5]), float(data[6]))
		self.rightColor = (float(data[7]), float(data[8]), float(data[9]), float(data[10]))
		if len(data) >= 12:
			self.blendFunc = int(data[11])
			if len(data) >= 13:
				self.colorType = int(data[12])
				if len(data) >= 14:
					self.leftColorType = int(data[13])
					if len(data) >= 15:
						self.rightColorType = int(data[14])

	def encode(self):
		"""Encode this to a byte array."""
		ret = []
		ret.append("%06f" % self.leftPosition)
		ret.append("%06f" % self.middlePosition)
		ret.append("%06f" % self.rightPosition)
		for chan in self.leftColor:
			ret.append("%06f" % chan)
		for chan in self.rightColor:
			ret.append("%06f" % chan)
		if self.blendFunc is not None:
			ret.append("%d" % self.blendFunc)
			if self.colorType is not None:
				ret.append("%d" % self.colorType)
				if self.leftColorType is not None:
					ret.append("%d" % self.leftColorType)
					if self.rightColorType is not None:
						ret.append("%d" % self.rightColorType)
		return " ".join(ret)

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		ret.append("Left Position: " + str(self.leftPosition))
		ret.append("Middle Position: " + str(self.middlePosition))
		ret.append("Right Position: " + str(self.rightPosition))
		ret.append("Left Color: " + str(self.leftColor))
		ret.append("Right Color: " + str(self.rightColor))
		ret.append("Blend Function: " + self.BLEND_FUNCTIONS[self.blendFunc])
		ret.append("Color Type: " + self.COLOR_TYPES[self.colorType])
		ret.append("Left Color Type: " + self.ENDPOINT_COLOR_TYPES[self.leftColorType])
		ret.append("Right Color Type: " + self.ENDPOINT_COLOR_TYPES[self.rightColorType])
		return ("\n" + indent).join(ret)


class GimpGgrGradient:
	"""Gimp color gradient.

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/ggr.txt
	"""

	def __init__(self, fileName: str | None = None):
		"""GimpGgrGradient.

		Args:
			fileName (str, optional): filename. Defaults to None.
		"""
		self.fileName = None
		self.segments = []
		self.name = ""
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		if isinstance(fileName, str):
			self.fileName = fileName
			file = open(fileName, "rb")
		else:
			self.fileName = fileName.name
			file = fileName
		data = file.read()
		file.close()
		self.decode(data)

	def decode(self, dataIn: bytes):
		"""Decode a byte buffer.

		Args:
			dataIn (bytes): data buffer to decode

		Raises:
			Exception: File format error.  Magic value mismatch.
		"""
		data = dataIn.decode("utf-8").split("\n")
		data = [l.strip() for l in data]
		if data[0] != "GIMP Gradient":
			raise Exception("File format error.  Magic value mismatch.")
		self.name = data[1].split(":", 1)[-1].strip()
		numSegments = int(data[2])
		for i in range(numSegments):
			gSeg = GradientSegment()
			gSeg.decode(data[i + 3])
			self.segments.append(gSeg)

	def encode(self):
		"""Encode this to a byte array."""
		ret = ["GIMP Gradient"]
		ret.append("Name: " + self.name)
		ret.append(str(len(self.segments)))
		for segment in self.segments:
			ret.append(segment.encode())
		return ("\n".join(ret) + "\n").encode("utf-8")

	def save(self, tofileName=None):
		"""Save this gimp image to a file."""
		if hasattr(tofileName, "write"):
			file = tofileName
		else:
			file = open(tofileName, "wb")
		file.write(self.encode())
		file.close()

	def getColor(self, percent):
		"""Given a decimal percent (1.0 = 100%) retrieve...

		the appropriate color for this point in the gradient.
		"""
		raise NotImplementedError()

	def __repr__(self, indent=""):
		"""Get a textual representation of this object."""
		ret = []
		if self.fileName is not None:
			ret.append("fileName: " + self.fileName)
		ret.append("Name: " + str(self.name))
		for seg in self.segments:
			ret.append(seg.__repr__(indent + "\t"))
		return ("\n" + indent).join(ret)



#!/usr/bin/env python3
"""Pure python implementation of the gimp xcf file format.

Currently supports:
	Loading xcf files
	Getting image hierarchy and info
	Getting image for each layer (PIL image)
Currently not supporting:
	Saving
	Programatically alter documents (add layer, etc)
	Rendering a final, compositied image
"""
from __future__ import annotations

import copy
from io import BytesIO, FileIO

import PIL.ImageGrab
from binaryiotools import IO
from blendmodes.blend import BlendType, blendLayers
from PIL import Image

from .GimpChannel import GimpChannel
from .GimpImageHierarchy import GimpImageHierarchy
from .GimpIOBase import GimpIOBase
from .GimpLayer import GimpLayer
from .GimpPrecision import Precision


class GimpDocument(GimpIOBase):
	"""Pure python implementation of the gimp file format.

	Has a series of attributes including the following:
	self._layers = None
	self._layerPtr = []
	self._channels = []
	self._channelPtr = []
	self.version = None
	self.width = 0
	self.height = 0
	self.baseColorMode = 0
	self.precision = None # Precision object
	self._data = None

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/xcf.txt
	"""

	def __init__(self, fileName=None):
		"""Pure python implementation of the gimp file format.

		Has a series of attributes including the following:
		self._layers = None
		self._layerPtr = []
		self._channels = []
		self._channelPtr = []
		self.version = None
		self.width = 0
		self.height = 0
		self.baseColorMode = 0
		self.precision = None # Precision object
		self._data = None

		See:
			https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/xcf.txt
		"""
		GimpIOBase.__init__(self, self)
		self._layers = []
		self._layerPtr = []
		self._channels = []
		self._channelPtr = []
		self.version = 11  # This is the most recent version
		self.width = 0
		self.height = 0
		self.baseColorMode = 0
		self.precision = None  # Precision object
		self._data = None
		self.fileName = None
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp xcf and decode the file. See decode for more on this process.

		:param fileName: can be a file name or a file-like object
		"""
		if isinstance(fileName, str):
			self.fileName = fileName
			file = open(fileName, "rb")
		else:
			self.fileName = fileName.name
			file = fileName
		data = file.read()
		file.close()
		self.decode(data)

	def decode(self, data: bytes, index: int = 0) -> int:
		"""Decode a byte buffer.

		Steps:
		Create a new IO buffer (array of binary values)
		Check that the file is a valid gimp xcf
		Grab the file version
		Grab other attributes as outlined in the spec
		Get precision data using the class and ioBuf buffer
		List of properties
		Get the layers and add the pointers to them
		Get the channels and add the pointers to them
		Return the offset

		Args:
			data (bytes): data buffer to decode
			index (int, optional): index within the buffer to start at]. Defaults to 0.

		Raises:
			Exception: "Not a valid GIMP file"

		Returns:
			int: offset
		"""
		# Create a new IO buffer (array of binary values)
		ioBuf = IO(data, index)
		# Check that the file is a valid gimp xcf
		if ioBuf.getBytes(9) != b"gimp xcf ":
			raise Exception("Not a valid GIMP file")
		# Grab the file version
		version = ioBuf.cString
		if version == "file":
			self.version = 0
		else:
			self.version = int(version[1:])
		# Grab other attributes as outlined in the spec
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.baseColorMode = ioBuf.u32
		# Get precision data using the class and ioBuf buffer
		self.precision = Precision()
		self.precision.decode(self.version, ioBuf)
		# List of properties
		self._propertiesDecode(ioBuf)
		self._layerPtr = []
		self._layers = []
		# Get the layers and add the pointers to them
		while True:
			ptr = self._pointerDecode(ioBuf)
			if ptr == 0:
				break
			self._layerPtr.append(ptr)
			layer = GimpLayer(self)
			layer.decode(ioBuf.data, ptr)
			self._layers.append(layer)
		# Get the channels and add the pointers to them
		self._channelPtr = []
		self._channels = []
		while True:
			ptr = self._pointerDecode(ioBuf)
			if ptr == 0:
				break
			self._channelPtr.append(ptr)
			chnl = GimpChannel(self)
			chnl.decode(ioBuf.data, ptr)
			self._channels.append(chnl)
		# Return the offset
		return ioBuf.index

	def encode(self):
		"""Encode to a byte array.

		Steps:
		Create a new IO buffer (array of binary values)
		The file is a valid gimp xcf
		Set the file version
		Set other attributes as outlined in the spec
		Set precision data using the class and ioBuf buffer
		List of properties
		Set the layers and add the pointers to them
		Set the channels and add the pointers to them
		Return the data

		"""
		# Create a new IO buffer (array of binary values)
		ioBuf = IO()
		# The file is a valid gimp xcf
		ioBuf.addBytes("gimp xcf ")
		# Set the file version
		ioBuf.addBytes(f"v{self.version:03d}" + "\0")
		# Set other attributes as outlined in the spec
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.u32 = self.baseColorMode
		# Set precision data using the class and ioBuf buffer
		if self.precision is None:
			self.precision = Precision()
		self.precision.encode(self.version, ioBuf)
		# List of properties
		ioBuf.addBytes(self._propertiesEncode())
		dataAreaIdx = ioBuf.index + self.POINTER_SIZE * (len(self.layers) + len(self._channels))
		dataAreaIO = IO()
		# Set the layers and add the pointers to them
		for layer in self.layers:
			ioBuf.index = dataAreaIdx + dataAreaIO.index
			dataAreaIO.addBytes(layer.encode())
		# Set the channels and add the pointers to them
		for channel in self._channels:
			ioBuf.index = dataAreaIdx + dataAreaIO.index
			dataAreaIO.addBytes(channel.encode())
		ioBuf.addBytes(dataAreaIO)
		# Return the data
		return ioBuf.data

	def forceFullyLoaded(self):
		"""Make sure everything is fully loaded from the file."""
		for layer in self.layers:
			layer.forceFullyLoaded()
		for channel in self._channels:
			channel.forceFullyLoaded()
		# no longer try to get the data from file
		self._layerPtr = None
		self._channelPtr = None
		self._data = None

	@property
	def layers(self):
		"""Decode the image's layers if necessary.

		TODO: need to do the same thing with self.Channels
		"""
		if len(self._layers) == 0:
			self._layers = []
			for ptr in self._layerPtr:
				layer = GimpLayer(self)
				layer.decode(self._data, ptr)
				self._layers.append(layer)
			# add a reference back to this object so it doesn't go away while array is in use
			self._layers.parent = self
			# override some internal methods so we can do more with them
			self._layers._actualDelitem_ = self._layers.__delitem__
			self._layers.__delitem__ = self.deleteLayer
			self._layers._actualSetitem_ = self._layers.__setitem__
			self._layers.__setitem__ = self.setLayer
		return self._layers

	def getLayer(self, index: int):
		"""Return a given layer."""
		return self.layers[index]

	def setLayer(self, index, layer):
		"""Assign to a given layer."""
		self.forceFullyLoaded()
		self._layerPtr = None  # no longer try to use the pointers to get data
		self.layers._actualSetitem_(index, layer)

	def newLayer(self, name: str, image: Image.Image, index: int = -1) -> GimpLayer:
		"""Create a new layer based on a PIL image.

		Args:
			name (str): a name for the new layer
			image (Image.Image): pil image
			index (int, optional): where to insert the new layer (default=top). Defaults to -1.

		Returns:
			GimpLayer: newly created GimpLayer object
		"""
		layer = GimpLayer(self, name, image)
		layer.imageHierarchy = GimpImageHierarchy(self, image=image)
		self.insertLayer(layer, index)
		return layer

	def newLayerFromClipboard(self, name: str = "pasted", index: int = -1) -> GimpLayer | None:
		"""Create a new image from the system clipboard.

		NOTE: requires pillow PIL implementation
		NOTE: only works on OSX and Windows

		Args:
			name (str): a name for the new layer
			index (int, optional): where to insert the new layer (default=top). Defaults to -1.

		Returns:
			GimpLayer: newly created GimpLayer object
		"""
		image = PIL.ImageGrab.grabclipboard()
		if isinstance(image, Image.Image):
			return self.newLayer(name, image, index)

	def addLayer(self, layer: GimpLayer):
		"""Append a layer object to the document.

		:param layer: the new layer to append
		"""
		self.insertLayer(layer, -1)

	def appendLayer(self, layer: GimpLayer):
		"""Append a layer object to the document.

		:param layer: the new layer to append
		"""
		self.insertLayer(layer, -1)

	def insertLayer(self, layer: GimpLayer, index: int = -1):
		"""Insert a layer object at a specific position.

		:param layer: the new layer to insert
		:param index: where to insert the new layer (default=top)
		"""
		self._layers.insert(index, layer)

	def deleteLayer(self, index: int) -> None:
		"""Delete a layer."""
		self.__delitem__(index)

	# make this class act like this class is an array of layers
	def __len__(self) -> int:
		"""Make this class act like this class is an array of layers...

		Get the len.
		"""
		return len(self.layers)

	def __getitem__(self, index: int) -> GimpLayer:
		"""Make this class act like this class is an array of layers...

		Get the layer at an index.
		"""
		return self.layers[index]

	def __setitem__(self, index: int, layer) -> None:
		"""Make this class act like this class is an array of layers...

		Set a layer at an index.
		"""
		self.setLayer(index, layer)

	def __delitem__(self, index: int) -> None:
		"""Make this class act like this class is an array of layers...

		Delete a layer at an index.
		"""
		self.deleteLayer(index)

	def __inc__(self, amt) -> GimpDocument:
		self.appendLayer(amt)
		return self

	@property
	def image(self):
		"""Get a final, compiled image."""
		# We need to preprocess the layers to sort them into a list of layers
		# and groups. Where a group is a list of layers

		# Example Layers [layer, layer, group, layer]
		# Example Group [Layer(Group), [layer, layer, ...]]
		layers = self.layers[:]  # Copy the attribute rather than write to it
		layersOut = []
		index = 0
		while index < len(layers):
			layerOrGroup = layers[index]
			if layerOrGroup.isGroup:
				elem = [layerOrGroup, []]
				index += 1
				while layers[index].itemPath is not None:
					layerCopy = copy.deepcopy(layers[index])
					layerCopy.xOffset -= layerOrGroup.xOffset
					layerCopy.yOffset -= layerOrGroup.yOffset
					elem[1].append(layerCopy)
					layers.pop(index)
				layersOut.append(elem)
			else:
				layersOut.append(layerOrGroup)
				index += 1
		return flattenAll(layersOut, (self.width, self.height))

	def save(self, filename: str | FileIO = None):
		"""Save this gimp image to a file."""
		self.forceFullyLoaded()
		if filename is None:
			filename = self.fileName
		if isinstance(filename, str):
			file = open(filename, "wb")
		else:
			file = filename
		file.write(self.encode())
		file.close()

	def saveNew(self, filename=None):
		"""Save a new gimp image to a file."""
		if filename is None:
			filename = self.fileName
		if isinstance(filename, str):
			file = open(filename, "wb")
		else:
			file = filename
		file.write(self.encode())
		file.close()

	def __repr__(self, indent="") -> str:
		"""Get a textual representation of this object."""
		del indent
		ret = []
		if self.fileName is not None:
			ret.append("fileName: " + self.fileName)
		ret.append("Version: " + str(self.version))
		ret.append("Size: " + str(self.width) + " x " + str(self.height))
		ret.append("Base Color Mode: " + self.COLOR_MODES[self.baseColorMode])
		ret.append("Precision: " + str(self.precision))
		ret.append(GimpIOBase.__repr__(self))
		if self._layerPtr:
			ret.append("Layers: ")
			for layer in self.layers:
				ret.append(layer.__repr__("\t"))
		if self._channelPtr:
			ret.append("Channels: ")
			for channel in self._channels:
				ret.append(channel.__repr__("\t"))
		return "\n".join(ret)


def blendModeLookup(
	blendmode: int, blendLookup: dict[int, BlendType], default: BlendType = BlendType.NORMAL
):
	"""Get the blendmode from a lookup table."""
	if blendmode not in blendLookup:
		print("WARNING " + str(blendmode) + " is not currently supported!")
		return default
	return blendLookup[blendmode]


def flattenLayerOrGroup(
	layerOrGroup: list[GimpLayer] | GimpLayer,
	imageDimensions: tuple[int, int],
	flattenedSoFar: Image.Image | None = None,
	ignoreHidden: bool = True,
) -> Image.Image:
	"""Flatten a layer or group on to an image of what has already been	flattened.

	Args:
		layerOrGroup (Layer,Group): A layer or a group of layers
		imageDimensions (tuple[int, int]): size of the image
		flattenedSoFar (PIL.Image, optional): the image of what has already
		been flattened. Defaults to None.
		ignoreHidden (bool, optional): ignore layers that are hidden. Defaults
		to True.

	Returns:
		PIL.Image: Flattened image
	"""
	blendLookup = {
		0: BlendType.NORMAL,
		3: BlendType.MULTIPLY,
		4: BlendType.SCREEN,
		5: BlendType.OVERLAY,
		6: BlendType.DIFFERENCE,
		7: BlendType.ADDITIVE,
		8: BlendType.NEGATION,
		9: BlendType.DARKEN,
		10: BlendType.LIGHTEN,
		11: BlendType.HUE,
		12: BlendType.SATURATION,
		13: BlendType.COLOUR,
		14: BlendType.LUMINOSITY,
		15: BlendType.DIVIDE,
		16: BlendType.COLOURDODGE,
		17: BlendType.COLOURBURN,
		18: BlendType.HARDLIGHT,
		19: BlendType.SOFTLIGHT,
		20: BlendType.GRAINEXTRACT,
		21: BlendType.GRAINMERGE,
		23: BlendType.OVERLAY,
		24: BlendType.HUE,
		25: BlendType.SATURATION,
		26: BlendType.COLOUR,
		27: BlendType.LUMINOSITY,
		28: BlendType.NORMAL,
		30: BlendType.MULTIPLY,
		31: BlendType.SCREEN,
		32: BlendType.DIFFERENCE,
		33: BlendType.ADDITIVE,
		34: BlendType.NEGATION,
		35: BlendType.DARKEN,
		36: BlendType.LIGHTEN,
		37: BlendType.HUE,
		38: BlendType.SATURATION,
		39: BlendType.COLOUR,
		40: BlendType.LUMINOSITY,
		41: BlendType.DIVIDE,
		42: BlendType.COLOURDODGE,
		43: BlendType.COLOURBURN,
		44: BlendType.HARDLIGHT,
		45: BlendType.SOFTLIGHT,
		46: BlendType.GRAINEXTRACT,
		47: BlendType.GRAINMERGE,
		48: BlendType.VIVIDLIGHT,
		49: BlendType.PINLIGHT,
		52: BlendType.EXCLUSION,
	}
	# Group
	if isinstance(layerOrGroup, list):
		if ignoreHidden and not layerOrGroup[0].visible:
			foregroundComposite = Image.new("RGBA", imageDimensions)
		else:  # A group is a list of layers
			# (see flattenAll)
			foregroundComposite = renderWOffset(
				flattenAll(layerOrGroup[1], imageDimensions, ignoreHidden),
				imageDimensions,
				(layerOrGroup[0].xOffset, layerOrGroup[0].yOffset),
			)
			if layerOrGroup[0].mask is not None:
				newFGComp = Image.new("RGBA", imageDimensions)
				newFGComp.paste(
					foregroundComposite,
					None,
					renderMaskWOffset(
						layerOrGroup[0].mask.image,
						imageDimensions,
						(layerOrGroup[0].xOffset, layerOrGroup[0].yOffset),
					),
				)
				foregroundComposite = newFGComp.convert("RGBA")
		if flattenedSoFar is None:
			return foregroundComposite
		return blendLayers(
			flattenedSoFar,
			foregroundComposite,
			blendModeLookup(layerOrGroup[0].blendMode, blendLookup),
			layerOrGroup[0].opacity,
		)

	# Layer
	if ignoreHidden and not layerOrGroup.visible:
		foregroundComposite = Image.new("RGBA", imageDimensions)
	else:
		# Get a raster image and apply blending
		foregroundComposite = renderWOffset(
			layerOrGroup.image, imageDimensions, (layerOrGroup.xOffset, layerOrGroup.yOffset)
		)
		if layerOrGroup.mask is not None:
			newFGComp = Image.new("RGBA", imageDimensions)
			newFGComp.paste(
				foregroundComposite,
				None,
				renderMaskWOffset(
					layerOrGroup.mask.image,
					imageDimensions,
					(layerOrGroup.xOffset, layerOrGroup.yOffset),
				),
			)
			foregroundComposite = newFGComp.convert("RGBA")
	if flattenedSoFar is None:
		return foregroundComposite

	return blendLayers(
		flattenedSoFar,
		foregroundComposite,
		blendModeLookup(layerOrGroup.blendMode, blendLookup),
		layerOrGroup.opacity,
	)


def flattenAll(
	layers: list[GimpLayer], imageDimensions: tuple[int, int], ignoreHidden: bool = True
) -> Image.Image:
	"""Flatten a list of layers and groups.

	Note the bottom layer is at the end of the list

	Args:
		layers (list[GimpLayer]): A list of layers and groups
		imageDimensions (tuple[int, int]): size of the image been flattened. Defaults to None.
		ignoreHidden (bool, optional): ignore layers that are hidden. Defaults
		to True.

	Returns:
		PIL.Image: Flattened image
	"""
	end = len(layers) - 1
	flattenedSoFar = flattenLayerOrGroup(layers[end], imageDimensions, ignoreHidden=ignoreHidden)
	for layer in range(end - 1, -1, -1):
		flattenedSoFar = flattenLayerOrGroup(
			layers[layer], imageDimensions, flattenedSoFar=flattenedSoFar, ignoreHidden=ignoreHidden
		)
	return flattenedSoFar


def renderWOffset(
	image: Image.Image, size: tuple[int, int], offsets: tuple[int, int] = (0, 0)
) -> Image.Image:
	"""Render an image with offset and alpha to a given size.

	Args:
		image (Image.Image): pil image to draw
		size (tuple[int, int]): width, height as a tuple
		alpha (float, optional): alpha transparency. Defaults to 1.0.
		offsets (tuple[int, int], optional): x, y offsets as a tuple.
		Defaults to (0, 0).

	Returns:
		Image.Image: new image
	"""
	imageOffset = Image.new("RGBA", size)
	imageOffset.paste(image.convert("RGBA"), offsets, image.convert("RGBA"))
	return imageOffset


def renderMaskWOffset(
	image: Image.Image, size: tuple[int, int], offsets: tuple[int, int] = (0, 0)
) -> Image.Image:
	"""Render an image with offset and alpha to a given size.

	Args:
		image (Image.Image): pil image to draw
		size (tuple[int, int]): width, height as a tuple
		alpha (float, optional): alpha transparency. Defaults to 1.0.
		offsets (tuple[int, int], optional): x, y offsets as a tuple.
		Defaults to (0, 0).

	Returns:
		Image.Image: new image
	"""
	mode = image.mode
	imageOffset = Image.new("RGBA", size)
	imageOffset.paste(image.convert("RGBA"), offsets, image.convert("RGBA"))
	return imageOffset.convert(mode)



"""Since the precision code is so unusual, I decided to create a class to parse it.
"""

# from .GimpImageHierarchy import GimpImageHierarchy
from __future__ import annotations

from binaryiotools import IO


class Precision:
	"""Since the precision code is so unusual, I decided to create a class to parse it."""

	def __init__(self):
		"""Since the precision code is so unusual, I decided to create a class to parse it."""
		self.bits = 8
		self.gamma = True
		self.numberFormat = int

	def decode(self, gimpVersion: int, ioBuf: IO):
		"""Decode the precision code from the file."""
		if gimpVersion < 4:
			self.bits = 8
			self.gamma = True
			self.numberFormat = int
		else:
			code = ioBuf.u32
			if gimpVersion == 4:
				self.gamma = (True, True, False, False, False)[code]
				self.bits = (8, 16, 32, 16, 32)[code]
				self.numberFormat = (int, int, int, float, float)[code]
			elif gimpVersion in (5, 6):
				self.gamma = code % 100 != 0
				code = int(code / 100)
				self.bits = (8, 16, 32, 16, 32)[code]
				self.numberFormat = (int, int, int, float, float)[code]
			else:  # gimpVersion 7 or above
				self.gamma = code % 100 != 0
				code = int(code / 100)
				self.bits = (8, 16, 32, 16, 32, 64)[code]
				self.numberFormat = (int, int, int, float, float, float)[code]

	def encode(self, gimpVersion: int, ioBuf: IO):
		"""Encode this to the file.

		NOTE: will not mess with development versions 5 or 6
		"""
		if gimpVersion < 4:
			if self.bits != 8 or not (self.gamma) or self.numberFormat != int:
				raise Exception(
					"Illegal precision (" + str(self) + ") for gimp version " + str(gimpVersion)
				)
		else:
			if gimpVersion == 4:
				if self.bits == 64:
					raise Exception(
						"Illegal precision (" + str(self) + ") for gimp version " + str(gimpVersion)
					)
				if self.numberFormat == int:
					code = (8, 16, 32).index(self.bits)
				else:
					code = (16, 32).index(self.bits) + 2
				code = code * 100
				if self.gamma:
					code += 50
			elif gimpVersion in (5, 6):
				raise NotImplementedError(
					"Cannot save to gimp developer version " + str(gimpVersion)
				)
			else:  # version 7 or above
				if self.numberFormat == int:
					code = (8, 16, 32).index(self.bits)
				else:
					code = (16, 32, 64).index(self.bits) + 2
				code = code * 100
				if self.gamma:
					code += 50
			ioBuf.u32 = code

	def requiredGimpVersion(self):
		"""Return the lowest gimp version that supports this precision."""
		if self.bits == 8 and self.gamma and self.numberFormat == int:
			return 0
		if self.bits == 64:
			return 7
		return 4

	def __repr__(self):
		"""Get a textual representation of this object."""
		ret = []
		ret.append(str(self.bits) + "-bit")
		ret.append("gamma" if self.gamma else "linear")
		ret.append("integer" if self.numberFormat is int else float)
		return " ".join(ret)



"""Represents a single channel or mask in a gimp image.
"""
from __future__ import annotations

from binaryiotools import IO
from PIL import Image

from .GimpImageHierarchy import GimpImageHierarchy
from .GimpIOBase import GimpIOBase


class GimpChannel(GimpIOBase):
	"""Represents a single channel or mask in a gimp image."""

	def __init__(self, parent, name: str = "", image: Image.Image | None = None):
		"""GimpChannel.

		Args:
			parent ([type]): some parent node/ object
			name (str, optional): name of the channel. Defaults to "".
			image (Image.Image, optional): image to set. Defaults to None.
		"""
		GimpIOBase.__init__(self, parent)
		self.width = 0
		self.height = 0
		self.name = name
		self._imageHierarchy = None
		self._imageHierarchyPtr = None
		if image is not None:  # this is last because image can reset values
			self.image = image
		self._data = None

	def decode(self, data: bytes, index: int = 0) -> int:
		"""Decode a byte buffer.

		Args:
			data (bytes): data to decode
			index (int, optional): index to start from. Defaults to 0.

		Returns:
			int: pointer
		"""
		ioBuf = IO(data, index)
		# print 'Decoding channel at',index
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.name = ioBuf.sz754
		self._propertiesDecode(ioBuf)
		self._imageHierarchyPtr = self._pointerDecode(ioBuf)
		self._data = ioBuf.data
		return ioBuf.index

	def encode(self) -> bytearray:
		"""Encode this object to a byte buffer."""
		ioBuf = IO()
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.sz754 = self.name
		ioBuf.addBytes(self._propertiesEncode())
		imgH = self._imageHierarchyPtr
		if imgH is None:
			imgH = 0
		ioBuf.addBytes(self._pointerEncode(imgH))
		return ioBuf.data

	@property
	def image(self) -> Image.Image | None:
		"""Get a final, compiled image."""
		return self.imageHierarchy.image

	@image.setter
	def image(self, image: Image.Image):
		"""Get a final, compiled image."""
		self.width = image.width
		self.height = image.height
		if not self.name and isinstance(image, str):
			# try to use a fileName as the name
			self.name = image.rsplit("\\", 1)[-1].rsplit("/", 1)[-1]
		self._imageHierarchy = GimpImageHierarchy(self, image)

	def forceFullyLoaded(self):
		"""Make sure everything is fully loaded from the file."""
		_ = self.image  # make sure the image is loaded so we can delete the hierarchy nonsense
		self._imageHierarchyPtr = None
		self._data = None

	@property
	def imageHierarchy(self):
		"""Get the image hierarchy.

		This is mainly used for decoding the image, so
		not much use to you.
		"""
		if self._imageHierarchy is None:
			self._imageHierarchy = GimpImageHierarchy(self)
			self._imageHierarchy.decode(self._data, self._imageHierarchyPtr)
		return self._imageHierarchy

	def __repr__(self, indent: str = "") -> str:
		"""Get a textual representation of this object."""
		ret = []
		ret.append("Name: " + str(self.name))
		ret.append("Size: " + str(self.width) + " x " + str(self.height))
		ret.append(GimpIOBase.__repr__(self, indent))
		return indent + (("\n" + indent).join(ret))



#!/usr/bin/env python3
"""Pure python implementation of the gimp vbr brush format.
"""
from __future__ import annotations

from io import BytesIO


class GimpVbrBrush:
	"""Pure python implementation of the gimp vbr brush format.

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/vbr.txt
	"""

	BRUSH_SHAPES = ["circle", "square", "diamond"]

	def __init__(self, fileName: BytesIO | str | None = None):
		"""Pure python implementation of the gimp vbr brush format.

		Args:
			fileName (BytesIO, str, optional): filename. Defaults to None.
		"""
		self.version = 1.0
		self.name = ""
		self.spacing = 0
		self.radius = 50
		self.hardness = 1
		self.aspectRatio = 1
		self.angle = 0
		self.brushShape = None  # one of the strings in self.BRUSH_SHAPES
		self.spikes = None
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		if isinstance(fileName, str):
			self.fileName = fileName
			file = open(fileName, "rb")
		else:
			self.fileName = fileName.name
			file = fileName
		data = file.read()
		file.close()
		self.decode(data)

	@property
	def image(self):
		"""This parametric brush converted to a useable PIL image."""
		raise NotImplementedError()  # TODO:

	def decode(self, dataIn: bytes):
		"""Decode a byte buffer.

		:param dataIn: data buffer to decode
		"""
		data = [s.strip() for s in dataIn.decode("utf-8").split("\n")]
		if data[0] != "GIMP-VBR":
			raise Exception("File format error.  Magic value mismatch.")
		self.version = float(data[1])
		if self.version == 1.0:
			self.name = data[2]  # max len 255 bytes
			self.spacing = float(data[3])
			self.radius = float(data[4])
			self.hardness = float(data[5])
			self.aspectRatio = float(data[6])
			self.angle = float(data[7])
		elif self.version == 1.5:
			self.name = data[2]  # max len 255 bytes
			self.brushShape = data[3]  # one of the strings in self.BRUSH_SHAPES
			self.spacing = float(data[4])
			self.radius = float(data[5])
			self.spikes = float(data[6])
			self.hardness = float(data[7])
			self.aspectRatio = float(data[8])
			self.angle = float(data[9])
		else:
			raise Exception("Unknown version " + str(self.version))

	def encode(self):
		"""Encode to a raw data stream."""
		data = []
		data.append("GIMP-VBR")
		data.append(str(self.version))
		if self.version == 1.0:
			data.append(str(self.name))
			data.append(str(self.spacing))
			data.append(str(self.radius))
			data.append(str(self.hardness))
			data.append(str(self.aspectRatio))
			data.append(str(self.angle))
		elif self.version == 1.5:
			data.append(str(self.name))
			data.append(str(self.brushShape))
			data.append(str(self.spacing))
			data.append(str(self.radius))
			data.append(str(self.spikes))
			data.append(str(self.hardness))
			data.append(str(self.aspectRatio))
			data.append(str(self.angle))
		return ("\n".join(data) + "\n").encode("utf-8")

	def save(self, tofileName=None, toExtension=None):
		"""Save this gimp image to a file."""
		asImage = False
		if toExtension is None:
			if tofileName is not None:
				toExtension = tofileName.rsplit(".", 1)
				if len(toExtension) > 1:
					toExtension = toExtension[-1]
				else:
					toExtension = None
		if toExtension is not None and toExtension != "vbr":
			asImage = True
		if asImage:
			self.image.save(tofileName)
		else:
			if hasattr(tofileName, "write"):
				file = tofileName
			else:
				file = open(tofileName, "wb")
			file.write(self.encode())
			file.close()

	def __repr__(self):
		"""Get a textual representation of this object."""
		ret = []
		if self.fileName is not None:
			ret.append("fileName: " + self.fileName)
		ret.append("Name: " + str(self.name))
		ret.append("Version: " + str(self.version))
		ret.append("Spacing: " + str(self.spacing))
		ret.append("Radius: " + str(self.radius))
		ret.append("Hardness: " + str(self.hardness))
		ret.append("Aspect ratio: " + str(self.aspectRatio))
		ret.append("Angle: " + str(self.angle))
		ret.append("Brush Shape: " + str(self.brushShape))
		ret.append("Spikes: " + str(self.spikes))
		return "\n".join(ret)

	def __eq__(self, other):
		"""Perform a comparison."""
		if other.name != self.name:
			return False
		if other.version != self.version:
			return False
		if other.spacing != self.spacing:
			return False
		if other.radius != self.radius:
			return False
		if other.hardness != self.hardness:
			return False
		if other.aspectRatio != self.aspectRatio:
			return False
		if other.angle != self.angle:
			return False
		if other.brushShape != self.brushShape:
			return False
		if other.spikes != self.spikes:
			return False
		return True



#!/usr/bin/env python3
"""Pure python implementation of a gimp pattern file.
"""
from __future__ import annotations

from io import BytesIO

import PIL.Image
from binaryiotools import IO


class GimpPatPattern:
	"""Pure python implementation of a gimp pattern file.

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/pat.txt
	"""

	COLOR_MODES = [None, "L", "LA", "RGB", "RGBA"]

	def __init__(self, fileName: BytesIO | str = None):
		"""Pure python implementation of a gimp pattern file.

		Args:
			fileName (BytesIO, optional): filename or pointer. Defaults to None.
		"""
		self.fileName = None
		self.version = 1
		self.width = 0
		self.height = 0
		self.bpp = 4
		self.mode = self.COLOR_MODES[self.bpp]
		self.name = ""
		self._rawImage = None
		self._image = None
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName: BytesIO | str):
		"""Load a gimp file.

		:param fileName: can be a file name or a file-like object
		"""
		if isinstance(fileName, str):
			self.fileName = fileName
			file = open(fileName, "rb")
		else:
			self.fileName = fileName.name
			file = fileName
		data = file.read()
		file.close()
		self.decode(data)

	def decode(self, data: bytes, index: int = 0):
		"""Decode a byte buffer.

		Args:
			data (bytes): data to decode
			index (int, optional): index to start from. Defaults to 0.

		Raises:
			Exception: "File format error.  Magic value mismatch."

		Returns:
			int: pointer
		"""
		ioBuf = IO(data, index)
		headerSize = ioBuf.u32
		self.version = ioBuf.u32
		self.width = ioBuf.u32
		self.height = ioBuf.u32
		self.bpp = ioBuf.u32
		self.mode = self.COLOR_MODES[self.bpp]
		magic = ioBuf.getBytes(4)
		if magic.decode("ascii") != "GPAT":
			raise Exception("File format error.  Magic value mismatch.")
		nameLen = headerSize - ioBuf.index
		self.name = ioBuf.getBytes(nameLen).decode("UTF-8")
		self._rawImage = ioBuf.getBytes(self.width * self.height * self.bpp)
		self._image = None
		return ioBuf.index

	def encode(self):
		"""Encode to a byte buffer."""
		ioBuf = IO()
		ioBuf.u32 = 24 + len(self.name)
		ioBuf.u32 = self.version
		ioBuf.u32 = self.width
		ioBuf.u32 = self.height
		ioBuf.u32 = len(self.image.mode)
		ioBuf.addBytes("GPAT")
		ioBuf.addBytes(self.name.encode("utf-8"))
		if self._rawImage is None:
			rawImage = self.image.tobytes(encoder_name="raw")
		else:
			rawImage = self._rawImage
		ioBuf.addBytes(rawImage)
		return ioBuf.data

	@property
	def size(self):
		"""The size of the pattern."""
		return (self.width, self.height)

	@property
	def image(self):
		"""Get a final, compiled image."""
		if self._image is None:
			if self._rawImage is None:
				return None
			self._image = PIL.Image.frombytes(
				self.mode, self.size, self._rawImage, decoder_name="raw"
			)
		return self._image

	@image.setter
	def image(self, image):
		self._image = image
		self._rawImage = None

	def save(self, tofileName=None, toExtension=None):
		"""Save this gimp image to a file."""
		asImage = False
		if toExtension is None:
			if tofileName is not None:
				toExtension = tofileName.rsplit(".", 1)
				if len(toExtension) > 1:
					toExtension = toExtension[-1]
				else:
					toExtension = None
		if toExtension is not None and toExtension != "pat":
			asImage = True
		if asImage:
			self.image.save(tofileName)
		else:
			if hasattr(tofileName, "write"):
				file = tofileName
			else:
				file = open(tofileName, "wb")
			file.write(self.encode())
			file.close()

	def __repr__(self):
		"""Get a textual representation of this object."""
		ret = []
		if self.fileName is not None:
			ret.append("fileName: " + self.fileName)
		ret.append("Name: " + str(self.name))
		ret.append("Version: " + str(self.version))
		ret.append("Size: " + str(self.width) + " x " + str(self.height))
		ret.append("BPP: " + str(self.bpp))
		ret.append("Mode: " + str(self.mode))
		return "\n".join(ret)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gimpformats']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.2.0,<9',
 'binaryiotools>=2020.2.2,<2022',
 'blendmodes>=2021.1,<2023',
 'brackettree>=0.2.5,<2',
 'numpy>=1.20.3,<2']

setup_kwargs = {
    'name': 'gimpformats',
    'version': '2021.1',
    'description': 'Pure python implementation of the gimp file format(s)',
    'long_description': '[![GitHub top language](https://img.shields.io/github/languages/top/FHPythonUtils/GimpFormats.svg?style=for-the-badge)](../../)\n[![Repository size](https://img.shields.io/github/repo-size/FHPythonUtils/GimpFormats.svg?style=for-the-badge)](../../)\n[![Issues](https://img.shields.io/github/issues/FHPythonUtils/GimpFormats.svg?style=for-the-badge)](../../issues)\n[![License](https://img.shields.io/github/license/FHPythonUtils/GimpFormats.svg?style=for-the-badge)](/LICENSE.md)\n[![Commit activity](https://img.shields.io/github/commit-activity/m/FHPythonUtils/GimpFormats.svg?style=for-the-badge)](../../commits/master)\n[![Last commit](https://img.shields.io/github/last-commit/FHPythonUtils/GimpFormats.svg?style=for-the-badge)](../../commits/master)\n[![PyPI Downloads](https://img.shields.io/pypi/dm/gimpformats.svg?style=for-the-badge)](https://pypistats.org/packages/gimpformats)\n[![PyPI Total Downloads](https://img.shields.io/badge/dynamic/json?style=for-the-badge&label=total%20downloads&query=%24.total_downloads&url=https%3A%2F%2Fapi.pepy.tech%2Fapi%2Fprojects%2Fgimpformats)](https://pepy.tech/project/gimpformats)\n[![PyPI Version](https://img.shields.io/pypi/v/gimpformats.svg?style=for-the-badge)](https://pypi.org/project/gimpformats)\n\n<!-- omit in TOC -->\n# GimpFormats\n\n<img src="readme-assets/icons/name.png" alt="Project Icon" width="750">\n\nForked from https://github.com/TheHeadlessSourceMan/gimpFormats\n\nA pure python implementation of the GIMP XCF image format. Use this to interact\nwith GIMP image formats\n\nIssues and contributions very much wanted/ needed :smile:\n\nPreviously under `gimpformats_unofficial` now under `gimpformats`\n\n- [Getting started](#getting-started)\n- [Next tasks (see below)](#next-tasks-see-below)\n- [Currently supports](#currently-supports)\n- [In progress but results in crashes and tests failing](#in-progress-but-results-in-crashes-and-tests-failing)\n- [Not implemented](#not-implemented)\n- [Documentation](#documentation)\n- [Install With PIP](#install-with-pip)\n- [Language information](#language-information)\n\t- [Built for](#built-for)\n- [Install Python on Windows](#install-python-on-windows)\n\t- [Chocolatey](#chocolatey)\n\t- [Download](#download)\n- [Install Python on Linux](#install-python-on-linux)\n\t- [Apt](#apt)\n- [How to run](#how-to-run)\n\t- [With VSCode](#with-vscode)\n\t- [From the Terminal](#from-the-terminal)\n- [Download Project](#download-project)\n\t- [Clone](#clone)\n\t\t- [Using The Command Line](#using-the-command-line)\n\t\t- [Using GitHub Desktop](#using-github-desktop)\n\t- [Download Zip File](#download-zip-file)\n- [Community Files](#community-files)\n\t- [Licence](#licence)\n\t- [Changelog](#changelog)\n\t- [Code of Conduct](#code-of-conduct)\n\t- [Contributing](#contributing)\n\t- [Security](#security)\n\t- [Support](#support)\n\t- [Rationale](#rationale)\n\n## Getting started\n\nRead an image\n```python\nfrom gimpformats.gimpXcfDocument import GimpDocument\nproject = GimpDocument("image.xcf")\n```\n\nIterate the image and report the contents of each group followed by the first\nlevel children of the image\n```python\n"""List data on groups followed by the direct children of a gimp xcf document\n"""\nlayers = project.layers\nindex = 0\nprint("## Group info...")\nwhile index < len(layers):\n\tlayerOrGroup = layers[index]\n\tif layerOrGroup.isGroup:\n\t\tindex += 1\n\t\twhile layers[index].itemPath is not None:\n\t\t\tprint("Group \\"" + layerOrGroup.name + "\\" contains Layer \\"" + layers[index].name + "\\"")\n\t\t\tlayers.pop(index)\n\telse:\n\t\tindex += 1\n\nprint("## Document direct children...")\nfor layerOrGroup in layers:\n\tprint("\\"" + layerOrGroup.name + "\\" is a " + ("Group" if layerOrGroup.isGroup else "Layer"))\n\n```\n\nExample output:\n\n```none\n## Group info...\nGroup "Layer Group" contains Layer "Layer"\nGroup "Layer Group" contains Layer "Layer2"\n## Document direct children...\n"bg #1" is a Layer\n"bg" is a Layer\n"bg #2" is a Layer\n"Transformation" is a Layer\n"Layer Group" is a Group\n"Background" is a Layer\n```\n\n## Next tasks (see below)\n- Saving\n\n## Currently supports\n- Loading xcf files (up to current GIMP version 2.10)\n- Getting image hierarchy and info\n- Getting image for each layer (PIL image)\n- .gbr brushes\n- .vbr brushes\n- .gpl palette files\n- .pat pattern files\n- .gtp tool presets\n- Generate a flattened image\n- Add new layers\n\n## In progress but results in crashes and tests failing\n- Saving\n- .ggr gradients - reads/saves fine, but I need to come up with a way to get the\nactual colours\n- .gih brush sets - BUG: seems to have more image data per brush than what\'s\nexpected\n- .gpb brush - should work, but I need some test files\n\n## Not implemented\n- Exported paths in .svg format. - Reading should be easy enough, but I need to\nensure I don\'t get a full-blown svg in the mix\n- Standard "parasites"\n\n## Documentation\nSee the [Docs](/DOCS/) for more information.\n\n## Install With PIP\n```python\npip install gimpformats\n```\n\nHead to https://pypi.org/project/gimpformats/ for more info\n\n## Language information\n### Built for\nThis program has been written for Python 3 and has been tested with\nPython version 3.9.0 <https://www.python.org/downloads/release/python-380/>.\n\n## Install Python on Windows\n### Chocolatey\n```powershell\nchoco install python\n```\n### Download\nTo install Python, go to <https://www.python.org/> and download the latest\nversion.\n\n## Install Python on Linux\n### Apt\n```bash\nsudo apt install python3.9\n```\n\n## How to run\n### With VSCode\n1. Open the .py file in vscode\n2. Ensure a python 3.9 interpreter is selected (Ctrl+Shift+P > Python:Select\nInterpreter > Python 3.9)\n3. Run by pressing Ctrl+F5 (if you are prompted to install any modules, accept)\n### From the Terminal\n```bash\n./[file].py\n```\n\n## Download Project\n### Clone\n#### Using The Command Line\n1. Press the Clone or download button in the top right\n2. Copy the URL (link)\n3. Open the command line and change directory to where you wish to\nclone to\n4. Type \'git clone\' followed by URL in step 2\n```bash\n$ git clone https://github.com/FHPythonUtils/GimpFormats\n```\n\nMore information can be found at\n<https://help.github.com/en/articles/cloning-a-repository>\n\n#### Using GitHub Desktop\n1. Press the Clone or download button in the top right\n2. Click open in desktop\n3. Choose the path for where you want and click Clone\n\nMore information can be found at\n<https://help.github.com/en/desktop/contributing-to-projects/cloning-a-repository-from-github-to-github-desktop>\n\n### Download Zip File\n\n1. Download this GitHub repository\n2. Extract the zip archive\n3. Copy/ move to the desired location\n\n## Community Files\n### Licence\nLGPLv3 License\n(See the [LICENSE](/LICENSE.txt) for more information.)\n\n### Changelog\nSee the [Changelog](/CHANGELOG.md) for more information.\n\n### Code of Conduct\nOnline communities include people from many backgrounds. The *Project*\ncontributors are committed to providing a friendly, safe and welcoming\nenvironment for all. Please see the\n[Code of Conduct](https://github.com/FHPythonUtils/.github/blob/master/CODE_OF_CONDUCT.md)\n for more information.\n\n### Contributing\nContributions are welcome, please see the\n[Contributing Guidelines](https://github.com/FHPythonUtils/.github/blob/master/CONTRIBUTING.md)\nfor more information.\n\n### Security\nThank you for improving the security of the project, please see the\n[Security Policy](https://github.com/FHPythonUtils/.github/blob/master/SECURITY.md)\nfor more information.\n\n### Support\nThank you for using this project, I hope it is of use to you. Please be aware that\nthose involved with the project often do so for fun along with other commitments\n(such as work, family, etc). Please see the\n[Support Policy](https://github.com/FHPythonUtils/.github/blob/master/SUPPORT.md)\nfor more information.\n\n### Rationale\nThe rationale acts as a guide to various processes regarding projects such as\nthe versioning scheme and the programming styles used. Please see the\n[Rationale](https://github.com/FHPythonUtils/.github/blob/master/RATIONALE.md)\nfor more information.\n',
    'author': 'FredHappyface',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/FHPythonUtils/GimpFormats',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
"""Parasites are arbitrary (meta)data strings that can be attached to a document tree item

They are used to store things like last-used plugin settings, gamma adjuetments, etc.

Format of known parasites:
	https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/parasites.txt
"""
from __future__ import annotations

from binaryiotools import IO

# TODO: how to best use these for our puproses??
KNOWN_DOCUMENT_PARASITES = [
	"jpeg-save-defaults",
	"png-save-defaults",
	"<plug-in>/_fu_data",
	"exif-orientation-rotate",
]
KNOWN_IMAGE_PARASITES = [
	"gimp-comment",
	"gimp-brush-name",
	"gimp-brush-pipe-name",
	"gimp-brush-pipe-parameters",
	"gimp-image-grid",
	"gimp-pattern-name",
	"tiff-save-options",
	"jpeg-save-options",
	"jpeg-exif-data",
	"jpeg-original-settings",
	"gamma",
	"chromaticity",
	"rendering-intent",
	"hot-spot",
	"exif-data",
	"gimp-metadata",
	"icc-profile",
	"icc-profile-name",
	"decompose-data",
	"print-settings",
	"print-page-setup",
	"dcm/XXXX-XXXX-AA",
]
KNOWN_LAYER_PARASITES = ["gimp-text-layer", "gfig"]


class GimpParasite:
	"""Parasites are arbitrary (meta)data strings that can be attached to a document tree item

	They are used to store things like last-used plugin settings, gamma adjuetments, etc.

	Format of known parasites:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/parasites.txt
	"""

	def __init__(self):
		self.name = ""
		self.flags = 0
		self.data = None

	def decode(self, data: bytes, index: int = 0) -> int:
		"""Decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		ioBuf = IO(data, index)
		self.name = ioBuf.sz754
		self.flags = ioBuf.u32
		dataLength = ioBuf.u32
		self.data = ioBuf.getBytes(dataLength)
		return ioBuf.index

	def encode(self):
		"""Encode a byte buffer

		:param data: data buffer to encode
		:param index: index within the buffer to start at
		"""
		ioBuf = IO()
		ioBuf.sz754 = self.name
		ioBuf.u32 = self.flags
		ioBuf.u32 = len(self.data)
		ioBuf.addBytes(self.data)
		return ioBuf.data

	def __repr__(self, indent: str = "") -> str:
		"""Get a textual representation of this object."""
		ret = []
		ret.append("Name: " + str(self.name))
		ret.append("Flags: " + str(self.flags))
		ret.append("Data Len: " + str(len(self.data)))
		return indent + (("\n" + indent).join(ret))
